// A class template for a sorted doubly-linked list.
#ifndef DOUBLYLINKEDLIST_H
#define DOUBLYLINKEDLIST_H

#include <iostream>
#include <cstdlib>
using namespace std;

//*********************************************
// The ListNode class creates a type used to  *
// store a node of the linked list.           *
//*********************************************

template <class T>
class ListNode
{
public:
	T value;           // Node value
	ListNode<T> *next; // Pointer to the next node
	ListNode<T> *previous; // Pointer to the previous node

	// Constructors
	ListNode ()
	{ next = NULL;
	  previous = NULL;
	}
	ListNode (T nodeValue)
	{ value = nodeValue;
	  next = NULL;
	  previous = NULL;
	}
};

//*********************************************
// DoublyLinkedList class                           *
//*********************************************

template <class T>
class DoublyLinkedList
{
private:
	ListNode<T> *head;   // List head pointer
	ListNode<T> *tail;   // List tail pointer
	int listSize;		 // count of items in the List

public:
	template <class T>
	class iterator {
	private:
		ListNode<T> * ptr;
	public:
		iterator() { ptr = NULL; }
		bool operator==(iterator<T> RHS) {
			return ptr==RHS.ptr;
		}
		bool operator!=(iterator<T> RHS) {
			return ptr != RHS.ptr;
		} 
		iterator<T> & operator=(iterator<T> RHS) {
			ptr = RHS.ptr;
			return *this;
		} // end operator++

		iterator<T> & operator++() {
			if (ptr->next == NULL) {
				std::cout << "ERROR operator++:  increment past the end\n";
				exit(-1);
			} else {
				ptr = ptr->next;
			} // end if
			return *this;
		} // end operator++

		iterator<T> & operator++(int) {
			if (ptr->next == NULL) {
				std::cout << "ERROR operator++:  increment past the end\n";
				exit(-1);
			} else {
				ptr = ptr->next;
			} // end if
			return *this;
		} // end operator++
		T operator*() {
			if (ptr == NULL || ptr->next == NULL || ptr->previous == NULL) {
				std::cout << "ERROR operator*:  ITERATOR INVALID\n";
				exit(-1);
			} else {
				return ptr->value;
			} // end if
		} // end operator*
	friend class DoublyLinkedList;
	}; // end iterator class

	// Constructor
	DoublyLinkedList();

	// Destructor
	~DoublyLinkedList();

	// Linked list operations
	void insertNode(T);
	bool deleteNode(T);	// deletes the specified value and return an indication of the delete was
						// successful or not
	void displayList() const;
	int size() const;
	iterator<T> begin() {
		iterator<T> iter;
		iter.ptr = head->next;
		return iter;
	} // end begin
	iterator<T> end() {
		iterator<T> iter;
		iter.ptr = tail;
		return iter;
	} // end 

};


//**************************************************
// Default constructor with header and trailer nodes *
//**************************************************

template <class T>
DoublyLinkedList<T>::DoublyLinkedList() {
	head = new ListNode<T>();
	tail = new ListNode<T>();
	head->next = tail;
	head->previous = NULL;
	tail->previous = head;
	tail->next = NULL;
	listSize = 0;
} // end Default constructor


//**************************************************
// displayList shows the value stored in each node *
// of the linked list pointed to by head.          *
//**************************************************

template <class T>
void DoublyLinkedList<T>::displayList() const
{
	ListNode<T> *nodePtr;  // To move through the list

	// Position nodePtr at the head of the list.
	nodePtr = head;

	// While nodePtr points to a node, traverse
	// the list.
	while (nodePtr)
	{
		// Display the value in this node.
		cout << nodePtr->value << endl;

		// Move to the next node.
		nodePtr = nodePtr->next;
	}
}

//**************************************************
// The insertNode function inserts a node with     *
// newValue into sorted order in the list.         *
//**************************************************

template <class T>
void DoublyLinkedList<T>::insertNode(T newValue)
{
	ListNode<T> *current;             // To traverse the list

	// Position current at the head of list.
	current = head->next;
	
	// Skip all nodes whose value is less than newValue.
	while (current != tail && current->value < newValue) {  
		current = current->next;
	} // end while

	// Allocate a new node and store newValue there.
	ListNode<T> *newNode = new ListNode<T>(newValue);

	// Link newNode into doubly-linked list
	newNode->next = current;
	newNode->previous = current->previous;
	current->previous = newNode;
	newNode->previous->next = newNode;
	listSize++;

}

//*****************************************************
// The deleteNode function searches for a node        *
// with searchValue as its value. The first node, if  *
// found, is deleted from the list and from memory.   *
//*****************************************************

template <class T>
bool DoublyLinkedList<T>::deleteNode(T searchValue)
{
	//ADD CODE HERE FOR PART A

} // end deleteNode

//**************************************************
// Destructor                                      *
// This function deletes every node in the list.   *
//**************************************************

template <class T>
DoublyLinkedList<T>::~DoublyLinkedList()
{
	ListNode<T> *nodePtr;   // To traverse the list
	ListNode<T> *nextNode;  // To point to the next node

	// Position nodePtr at the head of the list.
	nodePtr = head;

	// While nodePtr is not at the end of the list...
	while (nodePtr != NULL)
	{
		// Save a pointer to the next node.
		nextNode = nodePtr->next;

		// Delete the current node.
		delete nodePtr;

		// Position nodePtr at the next node.
		nodePtr = nextNode;
	}
}

template <class T>
int DoublyLinkedList<T>::size() const {
	return listSize;
} // end size

#endif
